"""
Project information.
"""

name = 'py-opengauss'
identity = 'http://github.com/vimiix/py-opengauss'

meaculpa = 'Python+openGauss'
abstract = 'Driver and tools library for openGauss'

version_info = (1, 3, 4)  # dev based on py-postgresql version 1.3.0
version = '.'.join(map(str, version_info))
