##
# .port
##
"""
Platform specific modules.

The subject of each module should be the feature and the target platform.
This is done to keep modules small and descriptive.

These modules are for internal use only.
"""
__docformat__ = 'reStructuredText'
